from collections import defaultdict, namedtuple
from itertools import tee
import re 

from Bio import Alphabet
from Bio.Seq import Seq  # need to clean this up
from hgvs.utils.altseq_to_hgvsp import AltSeqToHgvsp

from .utils import GenomePosition, GenomeIntervalTree, vcf_alt_affected_range

_VariantTreeTranscriptRecord = namedtuple("VariantTreeTranscriptRecord", [
    "feat", "five_prime_cds_feat", "three_prime_cds_feat",
    "five_prime_utr_feats", "three_prime_utr_feats", "translated_feats"
])

_VariantTreeCodingRecord = namedtuple("VariantTreeCodingRecord",
                                      ["transcript", "feat"])

_LocalizedSeq = namedtuple("LocalizedSeq", ["seq", "pos"])

# Create a mock `RefTranscriptData` type which behaves sufficiently similarly
# for the purposes of this module. Used to avoid the initialization step of the
# actual type.
_MockRefTranscriptData = namedtuple("RefTranscriptData", [
    "transcript_sequence", "aa_sequence", "cds_start", "cds_stop",
    "protein_accession"
])

# Create a mock `AltTranscriptData` type which behaves sufficiently similarly
# for the purposes of this module. Used to avoid the initialization step of the
# actual type.
_MockAltTranscriptData = namedtuple("AltTranscriptData", [
    "transcript_sequence", "aa_sequence", "cds_start", "cds_stop",
    "protein_accession", "is_frameshift", "variant_start_aa",
    "frameshift_start", "is_substitution", "is_ambiguous"
])

ProteinVariantResult = namedtuple(
    "ProteinVariantResult",
    ["query_variant", "predicted_variant", "transcript_feat"])


class ProteinVariantPredictor():
    # TODO: Now I think it would be nice if this was written to use GFF3
    # instead of GTF because of the hierarchic feature structure GFF3 provides,
    # which is is already being emulated (to an extent) below.
    def __init__(self, annotation_genome_tree, genome_faidx):
        self.genome_fasta = genome_faidx

        transcript_feats_dict = defaultdict(lambda: defaultdict(list))
        for feature in annotation_genome_tree.records:
            if feature.attributes.get("transcript_type") != "protein_coding":
                continue

            feature_type = feature.type
            if feature_type not in [
                "transcript", "CDS", "UTR", "start_codon", "stop_codon"
            ]:
                continue

            transcript_id = feature.attributes["transcript_id"]
            transcript_feats_dict[transcript_id][feature_type].append(feature)

        transcript_records = {}
        cds_records = []
        for transcript_id, features in transcript_feats_dict.items():
            # There should only be one transcript record...
            transcript_feat = features["transcript"][0]

            # Hopefully redundant check
            if not transcript_feat:
                print(f"Unexpected missing transcript feature for transcript"
                      " ID {transcript_id}")
                continue

            cds_feats = features["CDS"]

            # Sort the CDS features so that the CDS adjacent to the 5' UTR is
            # first and the CDS adjacent to the 3' UTR is last.
            is_forward_stranded = transcript_feat.is_forward_stranded
            cds_feats.sort(reverse=(not is_forward_stranded),
                           key=lambda feat: feat.pos.start
                           if is_forward_stranded else feat.pos.end)

            five_prime_cds_feat = cds_feats[0]
            three_prime_cds_feat = cds_feats[-1]

            utr_feats = features["UTR"]

            is_before = lambda rh: lambda lh: lh.pos.end <= rh.pos.start
            is_after = lambda rh: lambda lh: is_before(lh)(rh)

            utr_filter_5p = is_before(
                five_prime_cds_feat) if is_forward_stranded else is_after(
                five_prime_cds_feat)
            utr_filter_3p = is_after(
                three_prime_cds_feat) if is_forward_stranded else is_before(
                three_prime_cds_feat)

            five_prime_utr_feats = filter(utr_filter_5p, utr_feats)
            three_prime_utr_feats = filter(utr_filter_3p, utr_feats)

            transcript_record = _VariantTreeTranscriptRecord(
                feat=transcript_feat,
                five_prime_cds_feat=five_prime_cds_feat,
                three_prime_cds_feat=three_prime_cds_feat,
                five_prime_utr_feats=five_prime_utr_feats,
                three_prime_utr_feats=three_prime_utr_feats,
                translated_feats=features["start_codon"] + cds_feats +
                                 features["stop_codon"])

            transcript_records[transcript_id] = transcript_record

            for cds_feat in cds_feats:
                cds_records.append(
                    _VariantTreeCodingRecord(transcript=transcript_record,
                                             feat=cds_feat))

        self.transcript_records = transcript_records
        self.tree = GenomeIntervalTree(lambda record: record.feat.pos,
                                       cds_records)

        if 'chrM' in self.genome_fasta.keys():
            self.genome_fasta_mito = genome_faidx["chrM"][:]
        else:
            self.genome_fasta_mito = None

    @classmethod
    def _merged_and_sorted_intersecting_intervals(cls, intervals):
        current_interval = None

        for interval in sorted(intervals):
            if current_interval is None:
                current_interval = interval
                continue

            i0, i1, c0, c1 = *interval, *current_interval

            if i0 > c1:  # i1 < c0 shouldn't be possible due to sort
                # This new interval doesn't overlap with the current one; yield
                # the old interval and rotate.
                yield current_interval
                current_interval = interval
                continue

            # This interval intersects with the current one; merge them
            current_interval = (min(i0, c0), max(i1, c1))

        if current_interval is not None:
            yield current_interval

    @classmethod
    def _splice_seq(cls, seq, intervals):
        ordered_intervals = cls._merged_and_sorted_intersecting_intervals(
            intervals)

        # Create empty `Seq` with same alphabet
        spliced_seq = seq[0:0]

        for interval in ordered_intervals:
            spliced_seq += seq[slice(*interval)]

        return spliced_seq

    def predict_for_vcf_record(self, vcf_record):

        def check_str(s):
            ''' check a string to see if it has non-sequence characters '''
            match = re.match("^[AGCTN]*$", s)
            return match is not None


        def handle_mito(_tx_pos):
            ''' handles this chrM wierdness '''
            mito_seq = None
            if self.genome_fasta_mito:
                mito_seq = Seq(self.genome_fasta_mito.seq[_tx_pos.start:_tx_pos.end], 
                            alphabet=Alphabet.generic_dna)
            return mito_seq


        variant_results = []
        record_pos = GenomePosition.from_vcf_record_pos(vcf_record)

        ref = vcf_record.REF
        for alt in vcf_record.ALT:
            # Create a GenomePosition representing the range affected by the
            # ALT sequence.
            affected_pos = record_pos.shifted_by(
                vcf_alt_affected_range(ref, alt))

            coding_overlaps = self.tree.get_all_overlaps(affected_pos)
            transcript_ids = set(
                record.transcript.feat.attributes["transcript_id"]
                for record in coding_overlaps)

            for tx_id in transcript_ids:
                transcript = self.transcript_records[tx_id]
                tx_pos = transcript.feat.pos
                ref_pos = record_pos.shifted_by(0, len(ref) - 1)

                # We want to make sure that the transcript sequence we pull
                # includes the REF so that we can cleanly replace it with the
                # ALT later.
                tx_pos = tx_pos.shifted_by(
                    min(0, ref_pos.start - tx_pos.start),
                    max(0, ref_pos.end - tx_pos.end))

                ref_tx_seq = Seq(self.genome_fasta["chr" + tx_pos.chrom]
                                 [tx_pos.start:tx_pos.end].seq,
                                 alphabet=Alphabet.generic_dna)

                if not check_str(str(ref_tx_seq)): # non-DNA seq chars
                    if tx_pos.chrom == 'M':
                        ref_tx_seq = handle_mito(tx_pos)
                        if not ref_tx_seq:
                            continue
                    else:   # this shouldn't happen; skip iter
                        continue 

                ref_slice = ref_pos.slice_within(tx_pos)

                # TODO: The '*' ALT indicates that the allele was missing.
                # Right now, this is treated as a deletion of the entire REF
                # sequence, but perhaps this should be handled differently...
                alt_value = '' if alt.value == '*' else alt.value

                alt_tx_seq = ref_tx_seq[:ref_slice.start] \
                             + alt_value \
                             + ref_tx_seq[ref_slice.stop:]

                alt_tx_seq_len_delta = len(alt_tx_seq) - len(ref_tx_seq)

                ref_splice_intervals, alt_splice_intervals = tee(
                    (feat.pos.slice_within(tx_pos).indices(len(tx_pos))[:2]
                     for feat in transcript.translated_feats), 2)

                record_pos_tx_slice = record_pos.slice_within(tx_pos)

                # NOTE: Because of the way the splicing interval offsets are
                # calculated, insertions should not affect the length of the
                # ALT coding sequence; e.g. an insertion which starts before
                # the 5' CDS will just shift the splicing intervals over by its
                # length.
                alt_splice_intervals = (
                    # If the variant comes after a feature's position, the
                    # interval doesn't need to be changed; if it does,
                    # the length difference between the ALT and REF is added.
                    # This applies individually for both the start and end of
                    # each interval.
                    (ival[0] + (0 if record_pos_tx_slice.start >= ival[0] else
                                alt_tx_seq_len_delta),
                     ival[1] + (0 if record_pos_tx_slice.start >= ival[1] else
                                alt_tx_seq_len_delta))
                    for ival in alt_splice_intervals)

                ref_splice_intervals = list(ref_splice_intervals)
                alt_splice_intervals = list(alt_splice_intervals)

                ref_coding_seq = self._splice_seq(ref_tx_seq,
                                                  ref_splice_intervals)
                alt_coding_seq = self._splice_seq(alt_tx_seq,
                                                  alt_splice_intervals)

                if transcript.feat.is_reverse_stranded:
                    ref_coding_seq = ref_coding_seq.reverse_complement()
                    alt_coding_seq = alt_coding_seq.reverse_complement()

                # TODO: For instances where start/stop codons have been
                # altered, consider splicing in UTRs up until start/stop
                # codons.

                n_term_padding = -(transcript.five_prime_cds_feat.phase % -3)

                ref_c_term_trim = (n_term_padding + len(ref_coding_seq)) % 3
                alt_c_term_trim = (n_term_padding + len(alt_coding_seq)) % 3
                ref_trim_slice = slice(0,
                                       len(ref_coding_seq) - ref_c_term_trim)
                alt_trim_slice = slice(0,
                                       len(alt_coding_seq) - alt_c_term_trim)

                ref_coding_seq = (
                                     'N' * n_term_padding) + ref_coding_seq[
                                     ref_trim_slice]
                alt_coding_seq = (
                                     'N' * n_term_padding) + alt_coding_seq[
                                     alt_trim_slice]

                ref_aa_seq = ref_coding_seq.translate()
                alt_aa_seq = alt_coding_seq.translate()

                protein_id = transcript.feat.attributes["protein_id"]

                # Alt transcript data flags

                is_frameshift = (len(ref_coding_seq) -
                                 len(alt_coding_seq)) % 3 != 0
                variant_start_aa = None
                is_substitution = False
                # The implementation for `is_ambiguous` used by `hgvs`
                # considers "ambiguity" to mean "more than one stop codon."
                is_ambiguous = alt_aa_seq.count('*') > 1

                min_coding_seq_length = min(len(ref_coding_seq),
                                            len(alt_coding_seq))
                for pos in range(min_coding_seq_length):
                    ref_na, alt_na = ref_coding_seq[pos], alt_coding_seq[pos]

                    if ref_na != alt_na:
                        variant_start_aa = pos // 3
                        break
                else:
                    if len(ref_coding_seq) == len(alt_coding_seq):
                        # The coding sequence was unchanged by the variant.
                        continue

                    variant_start_aa = min_coding_seq_length // 3

                # NOTE: Not sure if `hgvs` expects `is_substitution` to be
                # evaluated on the amino acid-level or the nucleic acid-level.
                # Currently evaluated on the amino acid-level; therefore silent
                # substitutions aren't taken into account.
                min_aa_seq_length = min(len(ref_aa_seq), len(alt_aa_seq))
                if len(ref_aa_seq) == len(alt_aa_seq):
                    for pos in range(min_aa_seq_length):
                        ref_aa, alt_aa = ref_aa_seq[pos], alt_aa_seq[pos]

                        if ref_aa != alt_aa:
                            if is_substitution:
                                is_substitution = False
                                break

                            is_substitution = True

                if not is_substitution and (
                    (len(ref_aa_seq) > len(alt_aa_seq)
                     and ref_aa_seq.startswith(alt_aa_seq)) or
                    (len(alt_aa_seq) > len(ref_aa_seq)
                     and alt_aa_seq.startswith(ref_aa_seq))):
                    # FIXME (by filing a bug report?):
                    # `hgvs` has a weird bug where deletions/insertions that do
                    # not cause any observable variation at the protein level
                    # cause an `IndexError` while attempting to find the first
                    # varying amino acid. These must be skipped for now.
                    continue

                ref_tx_data = _MockRefTranscriptData(
                    transcript_sequence=str(ref_coding_seq),
                    aa_sequence=str(ref_aa_seq),
                    cds_start=None,  # Only used for initialization
                    cds_stop=None,  # Only used for initialization
                    protein_accession=protein_id)
                alt_tx_data = _MockAltTranscriptData(
                    transcript_sequence=str(alt_coding_seq),
                    aa_sequence=str(alt_aa_seq),
                    cds_start=None,  # Only used for initialization
                    cds_stop=None,  # Only used for initialization
                    protein_accession=protein_id,
                    is_frameshift=is_frameshift,
                    variant_start_aa=variant_start_aa + 1,  # 1-indexed
                    frameshift_start=None,  # Seems to be unused currently
                    is_substitution=is_substitution,
                    # The `AltTranscriptData` builder used by `hgvs` sets
                    # `is_ambiguous` to true if there are multiple stop codons
                    # in the AA sequence. This doesn't work in this code's
                    # context, because frameshifts typically create many new
                    # stop codons. It is uncertain why `hgvs` doesn't run into
                    # this issue; perhaps it's because it indiscriminately pads
                    # sequences that can't be cleanly translated.
                    is_ambiguous=(not is_frameshift) and is_ambiguous)

                protein_variant_builder = AltSeqToHgvsp(
                    ref_tx_data, alt_tx_data)

                protein_variant = protein_variant_builder.build_hgvsp()

                variant_results.append(
                    ProteinVariantResult(query_variant=alt,
                                         predicted_variant=protein_variant,
                                         transcript_feat=transcript.feat))

        return variant_results



import re
from collections import defaultdict
from pathlib import Path
import multiprocessing
from os import system 

import click
import hgvs.parser
import pandas as pd
import vcfpy
from hgvs.sequencevariant import SequenceVariant
from pathos.pools import _ProcessPool as Pool
from pyfaidx import Fasta
from tqdm import tqdm

from .protein_variant_predictor import ProteinVariantPredictor
from .utils import GenomePosition, GenomeIntervalTree, GFFFeature, \
    sequence_variants_are_equivalent


class AminoAcidMutationFinder():
    def __init__(self, cosmic_df, annotation_df, genome_faidx, cov_bool):

        if cosmic_df is not None:
            filtered_cosmic_df = self._make_filtered_cosmic_df(cosmic_df)

            self._cosmic_genome_tree = GenomeIntervalTree(
                lambda row: GenomePosition.from_str(
                    str(row["Mutation genome position"])),
                (row for _, row in filtered_cosmic_df.iterrows()))
        else:
            self._cosmic_genome_tree = None

        self._annotation_genome_tree = GenomeIntervalTree(
            lambda feat: feat.pos,
            (GFFFeature(row) for _, row in annotation_df.iterrows()))

        self._protein_variant_predictor = ProteinVariantPredictor(
            self._annotation_genome_tree, genome_faidx)

        self._coverage_bool = cov_bool

    @classmethod
    def _make_filtered_cosmic_df(cls, cosmic_df):
        """Return a view of the input `cosmic_df` filtered for
        `Primary site` == `lung`."""

        filtered_db = cosmic_df.loc[cosmic_df["Primary site"] == "lung"]
        return filtered_db

    hgvs_parser = hgvs.parser.Parser(expose_all_rules=True)
    mutation_aa_silent_sub_fix_pattern = re.compile(
        r"(p\.)([A-Z](?:[a-z]{2})?)(\d+)\2")

    def _get_cosmic_record_protein_variant(self, cosmic_record):
        mutation_aa = cosmic_record["Mutation AA"]
        transcript_accession = cosmic_record["Accession Number"]

        for tx_id, tx_record in self._protein_variant_predictor \
            .transcript_records.items():
            if tx_id.split('.')[0] == transcript_accession:
                transcript_record = tx_record
                break
        else:
            # If we can't find the transcript record, it probably is not
            # protein-coding.
            print(f"[{transcript_accession} -> N/A] {mutation_aa}")
            return None

        protein_accession = transcript_record.feat.attributes["protein_id"]

        # COSMIC incorrectly reports silent substitutions as `X{pos}X`, rather
        # than `X{pos}=`.
        mutation_aa = self.mutation_aa_silent_sub_fix_pattern.sub(
            r"\1\2\3=", mutation_aa)

        # COSMIC mutation CDS strings have uncertainty variants which are
        # not HGVS-compliant.

        cosmic_protein_posedit = (
            self.hgvs_parser  # pylint: disable=no-member
                .parse_p_typed_posedit(mutation_aa)).posedit

        cosmic_protein_variant = SequenceVariant(
            ac=protein_accession, type='p', posedit=cosmic_protein_posedit)

        return cosmic_protein_variant

    def _make_mutation_counts_df(self, cell_genemuts_pairs):
        """Transform a list of tuples in the form
        [(cell_name, {gene_name: {aa_mutation}})] into a `DataFrame`."""

        cell_names, cell_gene_muts = list(zip(*cell_genemuts_pairs))
        cell_names, cell_gene_muts = list(cell_names), list(cell_gene_muts)

        # Aggregate all of the gene names into a 2D array
        cell_gene_names = [
            list(gene_muts.keys()) for gene_muts in cell_gene_muts
        ]

        # Flatten them into a set
        all_gene_names = set([
            gene_name for sublist in cell_gene_names for gene_name in sublist
        ])

        mutation_data = {}
        for gene_name in all_gene_names:
            aa_mutations = [list() for _ in range(len(cell_genemuts_pairs))]

            for idx, gene_muts in enumerate(cell_gene_muts):
                aa_mutations[idx] = list(gene_muts[gene_name])

            mutation_data[gene_name] = aa_mutations

        # TODO: cell_names index?
        return pd.DataFrame(index=cell_names, data=mutation_data)

    def find_cell_gene_aa_mutations(self, stream=None, path=None):
        """Create a `dict` mapping gene names to amino acid-level mutations
        found in that gene (filtered by COSMIC).
        Accepts a `stream` or a `path` as input."""

        def extract_coverage(record_):
            """ extracts coverage info from the INFO field of a vcf entry """
            call = record_.calls[0]
            curr_AD = call.data.get('AD')

            if len(curr_AD) != 1:
                wt_count = curr_AD[0]
                v_count = curr_AD[1]  # need to account for > 1 alt allele
            else:
                wt_count = curr_AD[0]
                v_count = 0
            ratio_str = '[' + str(v_count) + ':' + str(wt_count) + ']'
            return (ratio_str)

        vcf_reader = vcfpy.Reader.from_stream(
            stream) if stream is not None else vcfpy.Reader.from_path(path)

        gene_aa_mutations = defaultdict(set)

        for record in vcf_reader:
            record_pos = GenomePosition.from_vcf_record(record)

            curr_cov = extract_coverage(record)

            # TODO: maybe clean this up into a method
            protein_variant_results = self._protein_variant_predictor \
                .predict_for_vcf_record(record)

            if not protein_variant_results:
                continue

            target_variants = None
            if self._cosmic_genome_tree:
                overlaps = self._cosmic_genome_tree.get_all_overlaps(
                    record_pos)

                target_variants = (
                    self._get_cosmic_record_protein_variant(overlap)
                    for overlap in overlaps)

                # Filter out variants which could not be correctly obtained for
                # some reason.
                target_variants = [
                    variant for variant in target_variants if variant
                ]

                if not target_variants:
                    continue

            for result in protein_variant_results:
                predicted_variant = result.predicted_variant

                if target_variants is not None:
                    for target_variant in target_variants:
                        # `strict_silent` is enabled because silent mutations
                        # have effects which are not discernable at the protein
                        # level and could easily be different at the DNA level.
                        if sequence_variants_are_equivalent(
                            target_variant,
                            predicted_variant,
                            strict_unknown=False,
                            strict_silent=True):
                            break
                    else:
                        # This protein variant didn't match any of the target
                        # variants; don't proceed.
                        continue

                # The predicted protein variant matches one or more target
                # variants (if there are any).
                if self._coverage_bool == 1:
                    pvr_str = str(predicted_variant) + ',' + curr_cov
                    gene_name = result.transcript_feat.attributes["gene_name"]
                    gene_aa_mutations[gene_name].add(pvr_str)
                else:
                    gene_name = result.transcript_feat.attributes["gene_name"]
                    gene_aa_mutations[gene_name].add(str(predicted_variant))

        return gene_aa_mutations

    def find_transcript_mutations(self, paths, processes=1):
        """Create a `DataFrame` of mutation counts, where the row indices are
        cell names and the column indices are gene names."""

        def init_process(aa_mutation_finder):
            global current_process_aa_mutation_finder
            current_process_aa_mutation_finder = aa_mutation_finder

        def process_cell(path):
            # print(path)
            return (
                Path(path).stem,
                current_process_aa_mutation_finder
                    .find_cell_gene_aa_mutations(path=path)
                    )

        if processes > 1:
            with Pool(processes, initializer=init_process,
                      initargs=(self,)) as pool:
                # results = tqdm(pool.map(process_cell, paths),
                # total=len(paths), smoothing=0.01)
                results = list(
                    tqdm(pool.imap(process_cell, paths),
                         total=len(paths),
                         smoothing=0.01))

        else:
            init_process(self)
            results = list(map(process_cell, tqdm(paths)))
            # results = list(map(process_cell, paths))  # testing

        return self._make_mutation_counts_df(results)


@click.command()
@click.option("--processes",
              "num_processes",
              default=1,
              help="number of processes to use for computation",
              type=int)
@click.option("--cosmicdb",
              "cosmicdb_path",
              help="optional path to cosmic db file (.tsv)",
              default=None)
@click.option("--annotation",
              "annotation_path",
              help="path to genome annotation (.gtf)",
              required=True)
@click.option("--genomefa",
              "genomefa_path",
              help="path to full genome sequences (.fasta)",
              required=True)
@click.option("--report_coverage",
              "cov_bool",
              help="do you want to report coverage information?",
              required=True, type=int)
@click.option("--output",
              "output_path",
              help="path to output file (.csv)",
              required=True)
@click.argument("input_files", required=True, nargs=-1)
def find_aa_mutations(num_processes, cosmicdb_path, annotation_path,
                      genomefa_path, cov_bool, output_path, input_files):
    """ report amino-acid level SNPs and indels in each sample, and
        associated coverage """

    print("Beginning setup (this may take several minutes!)")

    if cosmicdb_path:
        print("Loading COSMIC database...")
        cosmic_df = pd.read_csv(cosmicdb_path, sep='\t')
    else:
        cosmic_df = None

    print("Loading genome annotation...")
    annotation_df = pd.read_csv(annotation_path, sep='\t', skiprows=5)

    print("Loading genome sequences...")
    genome_faidx = Fasta(genomefa_path)

    print("Building genome trees...")
    aa_mutation_finder = AminoAcidMutationFinder(cosmic_df, annotation_df,
                                                genome_faidx, cov_bool)

    print("Setup complete.")

    print("Finding mutations...")
    result_df = aa_mutation_finder.find_transcript_mutations(input_files,
                                                processes=num_processes)
    print("Writing file...")
    output_path = Path(output_path)
    result_df.to_csv(output_path)
    # result_df.to_json(output_path)

    print("Done!")



import re

import numpy as np
import vcfpy
from hgvs import edit
from ncls import NCLS


class GenomePosition():
    genome_pos_pattern = re.compile(r"(.+):(\d+)-(\d+)")

    def __init__(self, chrom, start, end):
        self.chrom = chrom
        self.start = start
        self.end = end

    @classmethod
    def from_str(cls, pos_str):
        match = cls.genome_pos_pattern.match(pos_str)

        if not match:
            return None

        return cls(match[1], int(match[2]) - 1, int(match[3]))

    @classmethod
    def from_vcf_record(cls, record):
        CHROM = record.CHROM.replace("chr", "")

        affected_ranges = [
            vcf_alt_affected_range(record.REF, alt) for alt in record.ALT
        ]
        start = record.begin + min(map(lambda r: r.start, affected_ranges),
                                   default=0)
        end = record.begin + max(map(lambda r: r.stop, affected_ranges),
                                 default=1)

        return cls(CHROM, start, end)

    @classmethod
    def from_vcf_record_pos(cls, record):
        CHROM = record.CHROM.replace("chr", "")
        return cls(CHROM, record.begin, record.begin + 1)

    @classmethod
    def from_gtf_record(cls, record):
        return cls(record[0].replace("chr", ""),
                   int(record[3]) - 1, int(record[4]))

    def __eq__(self, other):
        return self.chrom == other.chrom and self.start == other.start and \
        self.end == other.end

    def __repr__(self):
        return "%s:%d-%d" % (self.chrom, self.start + 1, self.end)

    def __str__(self):
        return "%s:%d-%d" % (self.chrom, self.start + 1, self.end)

    def __len__(self):
        return self.end - self.start

    def __contains__(self, other):
        same_chrom = other.chrom == self.chrom
        same_start = other.start >= self.start
        same_end = other.end <= self.end
        return same_chrom and same_start and same_end

    def __and__(self, other):
        if self.chrom != other.chrom:
            return None

        if other.start >= self.end or self.start >= other.end:
            return None

        return self.__class__(self.chrom, max(self.start, other.start),
                              min(self.end, other.end))

    # FIXME: This method may be too overloaded...
    def shifted_by(self, start, end=None):
        if isinstance(start, range):
            start, end = start.start, start.stop

        if end is None:
            end = start

        return self.__class__(self.chrom, self.start + start, self.end + end)

    def slice_within(self, other):
        if self.chrom != other.chrom:
            return None

        if self.start < other.start or self.end > other.end:
            return None

        return slice(self.start - other.start, self.end - other.start)


class GenomeIntervalTree():
    def __init__(self, predicate, records):
        self.predicate = predicate
        self.records = []

        working_tree_map = {}

        idx = 0
        for record in records:
            genome_pos = predicate(record)

            if genome_pos is None:
                continue

            chrom = genome_pos.chrom

            if chrom not in working_tree_map:
                # (starts, ends, ids)
                working_tree_map[chrom] = ([], [], [])

            starts, ends, ids = working_tree_map[chrom]
            starts.append(genome_pos.start)
            ends.append(genome_pos.end)
            ids.append(idx)

            self.records.append(record)
            idx += 1

        tree_map = {}

        for chrom, (starts, ends, ids) in working_tree_map.items():
            tree_map[chrom] = NCLS(np.array(starts, dtype=np.long),
                                   np.array(ends, dtype=np.long),
                                   np.array(ids, dtype=np.long))

        self.tree_map = tree_map

    def _intervals(self, chrom):
        return self.tree_map[chrom].intervals()

    def _make_query_params(self, genome_pos_list):
        starts = np.array([genome_pos.start for genome_pos in genome_pos_list])
        ends = np.array([genome_pos.end for genome_pos in genome_pos_list])
        ids = np.array(list(range(len(genome_pos_list))))

        return (starts, ends, ids)

    def _pick_best_record(self, from_ids=None, for_pos=None):
        if len(from_ids) < 1:
            return None

        if len(from_ids) == 1:
            return self.records[from_ids[0]]

        records = [self.records[record_id] for record_id in from_ids]

        scored_records = [(record,
                           self._compute_jaccard_index(for_pos,
                                                       self.predicate(record)))
                          for record in records]
        sorted_records = sorted(scored_records,
                                key=lambda tup: tup[1],
                                reverse=True)

        return sorted_records[0][0]

    def _compute_jaccard_index(self, pos_a, pos_b):
        intersection = pos_a & pos_b

        if not intersection:
            return 0

        # The following is equivalent to |A ∩ B| / |A ∪ B|, but avoids
        # computing a union.
        # |A ∩ B| / (|A| + |B| - |A ∩ B|)
        return len(intersection) / (len(pos_a) + len(pos_b) -
                                    len(intersection))

    def has_overlap(self, genome_pos):
        tree = self.tree_map.get(genome_pos.chrom)

        if not tree:
            return False

        return tree.has_overlap(genome_pos.start, genome_pos.end)

    def get_first_overlap(self, genome_pos):
        tree = self.tree_map.get(genome_pos.chrom)

        if not tree:
            return None

        qparams = self._make_query_params([genome_pos])
        _, record_ids = tree.first_overlap_both(*qparams)

        if len(record_ids) < 1:
            return None

        return self.records[record_ids[0]]

    def get_best_overlap(self, genome_pos):
        tree = self.tree_map.get(genome_pos.chrom)

        if not tree:
            return None

        qparams = self._make_query_params([genome_pos])
        _, record_ids = tree.all_overlaps_both(*qparams)

        return self._pick_best_record(from_ids=record_ids, for_pos=genome_pos)

    def get_all_overlaps(self, genome_pos):
        tree = self.tree_map.get(genome_pos.chrom)

        if not tree:
            return []

        qparams = self._make_query_params([genome_pos])
        _, record_ids = tree.all_overlaps_both(*qparams)

        if any(map(lambda r: r >= len(self.records), record_ids)):
            print("uh oh i dont like this")
            print(len(self.records))
            print(record_ids)

        return [self.records[record_id] for record_id in record_ids]

    def get_first_containment(self, genome_pos):
        tree = self.tree_map.get(genome_pos.chrom)

        if not tree:
            return None

        qparams = self._make_query_params([genome_pos])
        _, record_ids = tree.all_containments_both(*qparams)

        if len(record_ids) < 1:
            return None

        return self.records[record_ids[0]]

    def get_best_containment(self, genome_pos):
        tree = self.tree_map.get(genome_pos.chrom)

        if not tree:
            return None

        qparams = self._make_query_params([genome_pos])
        _, record_ids = tree.all_containments_both(*qparams)

        return self._pick_best_record(from_ids=record_ids, for_pos=genome_pos)

    def get_all_containments(self, genome_pos):
        tree = self.tree_map.get(genome_pos.chrom)

        if not tree:
            return []

        qparams = self._make_query_params([genome_pos])
        _, record_ids = tree.all_containments_both(*qparams)

        return [self.records[record_id] for record_id in record_ids]


class GFFFeature():
    @classmethod
    def parse_gff_attributes(cls, attr_str):
        attr_dict = {}
        for key, value in (kv_str.split(' ')
                           for kv_str in re.split('; ?', attr_str) if kv_str):
            if '"' in value:
                value = value[1:-1]
            else:
                value = int(value)

            attr_dict[key] = value

        return attr_dict

    @property
    def is_forward_stranded(self):
        return self.strand == '+'

    @property
    def is_reverse_stranded(self):
        return self.strand == '-'

    def __init__(self, record):
        self.pos = GenomePosition.from_gtf_record(record)
        self.source = record[1]
        self.type = record[2]
        self.score = None if record[5] == '.' else float(record[5])
        self.strand = record[6]
        self.phase = None if record[7] == '.' else int(record[7])
        self.attributes = self.parse_gff_attributes(record[8])


def vcf_alt_affected_range(ref, alt):
    # TODO: This method currently only deals with simple substitutions.

    if alt.type in [vcfpy.SNV, vcfpy.MNV]:
        return range(len(ref))
    elif alt.type == vcfpy.INS:
        return range(2)
    elif alt.type == vcfpy.DEL:
        return range(1, len(ref))
    elif alt.type == vcfpy.INDEL:
        return range(len(ref))

    raise NotImplementedError()


def _seqs_are_equal(seq_a, seq_b, wildcard=None):
    if not len(seq_a) == len(seq_b):
        return False

    for a, b in zip(seq_a, seq_b):
        if a == wildcard or b == wildcard:
            continue

        if not a == b:
            return False

    return True


# This could be extended for other types of `SequenceVariant`s in the future if
# needed.
def sequence_variants_are_equivalent(seqvar_a,
                                     seqvar_b,
                                     strict_uncertain=False,
                                     strict_unknown=True,
                                     strict_silent=False):
    """Check if `seqvar_a` and `seqvar_b` are equivalent.
    Currently only works correctly for protein-level variants.

    Parameters
    ---------
    strict_uncertain : bool
        True if variant (position/edit) uncertainty is factored into
        this equivalency check. (default False)
    strict_unknown : bool
        True if unknown sequence units (e.g. 'X' for amino acids) should
        not match known sequence units. (default True)
    strict_silent : bool
        True if synonymous variants (e.g. 'Arg17=') should not match
        otherwise equivalent variants. (default False)
    """

    if not seqvar_a.ac == seqvar_b.ac:
        return False

    if not seqvar_a.type == seqvar_b.type:
        return False

    sv_type = seqvar_a.type

    if sv_type not in ["p"]:
        raise NotImplementedError()

    posedit_a, posedit_b = seqvar_a.posedit, seqvar_b.posedit

    if (posedit_a is None) or (posedit_b is None):
        return posedit_a is None and posedit_b is None

    if strict_uncertain and not posedit_a.uncertain == posedit_b.uncertain:
        return False

    pos_a, pos_b = posedit_a.pos, posedit_b.pos

    # TODO: Handle positional uncertainty

    if not pos_a == pos_b:
        return False

    edit_a, edit_b = posedit_a.edit, posedit_b.edit

    if not type(edit_a) is type(edit_b):
        print(type(edit_a), type(edit_b))
        return False

    _seqs_cmp = lambda a, b: _seqs_are_equal(
        a, b, wildcard=(None if strict_unknown else 'X'))

    if isinstance(edit_a, (edit.AARefAlt, edit.AAFs, edit.AAExt)):
        if (edit_a is None) or (edit_b is None):
            return edit_a is None and edit_b is None

        if not _seqs_cmp(edit_a.ref, edit_b.ref):
            return False

        if not _seqs_cmp(edit_a.alt, edit_b.alt):
            return False

        if strict_silent and (not edit_a.ref) and (not edit_a.alt):
            return False
    else:
        raise NotImplementedError()

    if isinstance(edit_a, (edit.AAFs, edit.AAExt)):
        if not edit_a.length == edit_b.length:
            return False

    if isinstance(edit_b, (edit.AAExt)):
        if not _seqs_cmp(edit_a.aaterm, edit_b.aaterm):
            return False

    return True



import os
import pandas as pd
import click
import vcfpy
from pathlib import Path
from tqdm import tqdm
from pathos.multiprocessing import _ProcessPool as Pool, ThreadPool

from .utils import GenomePosition, GenomeIntervalTree


def create_germline_genome_tree(germline_vcf_paths):
    """Create a `GenomeIntervalTree` comprised of the VCF records supplied by
    `germline_vcf_paths`."""

    germline_vcf_records = [list(vcfpy.Reader.from_path(path)) for path in
                            germline_vcf_paths]

    # Flatten records
    germline_vcf_records = sum(germline_vcf_records, [])

    return GenomeIntervalTree(GenomePosition.from_vcf_record,
                              germline_vcf_records)


def write_filtered_vcf(cell_vcf_stream, germline_tree, out_stream):
    """Rewrite `cell_vcf_stream` to `out_stream` while removing any mutations
    found in dbSNP or shared with the VCF records in `germline_tree`."""

    cell_vcf = vcfpy.Reader.from_stream(cell_vcf_stream)
    out_vcf = vcfpy.Writer.from_stream(out_stream, header=cell_vcf.header)

    for record in cell_vcf:
        # If a record's ID field is `.`, that means that the calling
        # software did not find an ID for it in the associated database,
        # typically dbSNP. This is represented as an empty array (`[]`)
        # in VCFPy.

        if record.ID:   # This record is in dbSNP; skip it.
            continue

        genome_pos = GenomePosition.from_vcf_record(record)

        # Filter for only containments before checking for record equality.
        # This avoids O(n*m) complexity.
        germline_records = germline_tree.get_all_containments(genome_pos)

        unique_alts = set(record.ALT)

        for germline_record in germline_records:
            if (
                germline_record.POS != record.POS or
                germline_record.REF != record.REF):
                continue

            unique_alts -= set(germline_record.ALT)

        # Write this record if there are still has remaining (unique) ALTs.
        if unique_alts:
            record.ALT = list(unique_alts)
            out_vcf.write_record(record)


""" launch """


@click.command()
@click.option("--processes", default=1,
              prompt="number of processes to use for computation", type=int)
@click.option("--germline", "germline_path",
              prompt="path to germline vcf files directory", required=True)
@click.option("--cells", "cells_path",
              prompt="path to cell vcf files directory", required=True)
@click.option("--metadata", "metadata_path",
              prompt="path to metadata csv file", required=True)
@click.option("--outdir", "out_path",
              prompt="path to output vcf files directory", required=True)
def germline_filter(processes, germline_path, cells_path, metadata_path,
                    out_path):
    """ filter out common SNPs/indels between germline samples and samples
        of interest """
    germline_path = Path(germline_path)
    cells_path = Path(cells_path)
    metadata_path = Path(metadata_path)
    out_path = Path(out_path)

    metadata_df = pd.read_csv(metadata_path)

    # Create a set of all patient IDs from the metadata file.
    all_patient_ids = set(metadata_df["patient_id"])

    def process_patient(patient_id):
        # Find all non-tumor bulk VCF files for the patient ID.
        germline_wb_vcf_paths = list(
            germline_path.glob(patient_id + "_*_*.vcf"))

        # Fetch all cell IDs associated with the patient ID.
        cell_ids = metadata_df.loc[metadata_df["patient_id"] == patient_id][
            "cell_id"]

        # Use the cell IDs to create a list of all single-cell VCF files
        # for the patient.
        cell_vcf_paths = [(cells_path / cell_id).with_suffix(".vcf") for
                          cell_id in cell_ids]

        # Create a genome interval tree for the patient's germline bulk VCF
        # data. Only selects one germline VCF to avoid over-filtering for
        # patients with multiple germline VCFs.
        germline_tree = create_germline_genome_tree(germline_wb_vcf_paths[0:1])

        def process_cell(cell_vcf_path):
            if not cell_vcf_path.exists():
                return

            # If there were any germline VCFs for this patient, append `GF_` to
            # the file name to indicate that the output VCF was
            # germline-filtered, not just dbSNP-filtered.
            out_name_prefix = "" if len(germline_wb_vcf_paths) < 1 else "GF_"
            out_vcf_path = out_path / (out_name_prefix + cell_vcf_path.name)

            with open(cell_vcf_path, mode='r') as in_file:
                with open(out_vcf_path, mode='w') as out_file:
                    write_filtered_vcf(in_file, germline_tree, out_file)

        # TODO: Maybe remove this in Python 3.8.
        # This thread pool max-worker count is from the implementation in
        # Python 3.8. Assuming that Pathos adopts the same semantics,
        # this can be removed.
        with ThreadPool(min(32, os.cpu_count() + 4)) as pool:
            pool.map(process_cell, cell_vcf_paths)

    print("Running germline filter...")
    if processes > 1:
        with Pool(processes) as pool:
            list(tqdm(pool.imap(process_patient, all_patient_ids),
                      total=len(all_patient_ids), smoothing=0.01))
    else:
        list(map(process_patient, tqdm(all_patient_ids, smoothing=0.1)))

    print("Done!")



import re
from pathlib import Path

import click
import numpy as np
import pandas as pd
import vcfpy
from pathos.pools import _ProcessPool as Pool
from tqdm import tqdm

from .utils import GenomePosition, GenomeIntervalTree


class MutationCounter():
    def __init__(self, cosmic_df, hg38_df):
        filtered_cosmic_df = self._make_filtered_cosmic_df(cosmic_df)

        self._cosmic_genome_tree = GenomeIntervalTree(
            lambda row: GenomePosition.from_str(
                str(row["Mutation genome position"])),
            (record for idx, record in filtered_cosmic_df.iterrows()))
        self._hg38_genome_tree = GenomeIntervalTree(
            GenomePosition.from_gtf_record,
            (record for idx, record in hg38_df.iterrows()))

    def _cosmic_subset_contains_genome_pos(self, genome_pos):
        return self._cosmic_genome_tree.has_overlap(genome_pos)

    def _find_containing_gene_record(self, genome_pos):
        return self._hg38_genome_tree.get_best_overlap(genome_pos)

    gene_name_pattern = re.compile(r"gene_name \"(.+?)\"")

    def _parse_gene_name(self, metadata):
        """Parse out a gene name from a GTF metadata string."""

        gene_name_match = self.gene_name_pattern.search(metadata)

        if not gene_name_match:
            return None

        return gene_name_match[1]

    def _make_filtered_cosmic_df(self, cosmic_df):
        """Return a view of the input `cosmic_df` filtered for
        `Primary site` == `lung`."""

        filtered_db = cosmic_df.loc[cosmic_df["Primary site"] == "lung"]
        return filtered_db

    def _make_mutation_counts_df(self, cell_genemuts_pairs):
        """Transform a list of tuples in the form
        (cell_name, {gene_name: mutation_count}) into a `DataFrame`."""

        cell_names, cell_gene_muts = list(zip(*cell_genemuts_pairs))
        cell_names, cell_gene_muts = list(cell_names), list(cell_gene_muts)

        # Aggregate all of the gene names into a 2D array
        cell_gene_names = [list(gene_muts.keys()) for gene_muts in
                           cell_gene_muts]
        # Flatten them into a set
        all_gene_names = set(
            [gene_name for sublist in cell_gene_names for gene_name in
             sublist])

        mutation_data = {}
        for gene_name in all_gene_names:
            mutation_counts = np.zeros(len(cell_genemuts_pairs), dtype=int)

            for idx, gene_muts in enumerate(cell_gene_muts):
                mutation_counts[idx] = gene_muts.get(gene_name, 0)

            mutation_data[gene_name] = mutation_counts

        # TODO: cell_names index?
        return pd.DataFrame(index=cell_names, data=mutation_data)

    def find_cell_gene_mut_counts(self, stream=None, path=None):
        """Create a `dict` mapping gene names to the number of times a mutation
        was found in that gene. Accepts a `stream` or a `path` as input."""

        vcf_reader = vcfpy.Reader.from_stream(
            stream) if stream is not None else vcfpy.Reader.from_path(path)

        gene_mutation_counts, filtered_gene_mutation_counts = {}, {}

        for record in vcf_reader:
            genome_pos = GenomePosition.from_vcf_record(record)
            # TODO: Filter out duplicates?
            # And is it better to filter out positional duplicates or gene name
            # duplicates?

            gene_row = self._find_containing_gene_record(genome_pos)

            if gene_row is None:
                continue

            gene_name = self._parse_gene_name(gene_row[8])

            if gene_name is None:
                continue

            gene_mutation_counts[gene_name] = gene_mutation_counts.get(
                gene_name, 0) + 1

            if not self._cosmic_subset_contains_genome_pos(genome_pos):
                continue

            filtered_gene_mutation_counts[
                gene_name] = filtered_gene_mutation_counts.get(gene_name,
                                                               0) + 1

        return gene_mutation_counts, filtered_gene_mutation_counts

    def find_mutation_counts(self, paths, processes=4):
        """Create a `DataFrame` of mutation counts, where the row indices are
        cell names and the column indices are gene names."""

        def init_process(mutation_counter):
            global current_process_mutation_counter
            current_process_mutation_counter = mutation_counter

        def process_cell(path):
            cell_name = Path(path).stem
            return (cell_name,
                    current_process_mutation_counter.find_cell_gene_mut_counts(
                        path=path))

        if processes > 1:
            with Pool(processes, initializer=init_process,
                      initargs=(self,)) as pool:
                results = list(
                    tqdm(pool.imap(process_cell, paths), total=len(paths),
                         smoothing=0.01))
        else:
            results = list(map(process_cell, tqdm(paths)))

        cell_genemuts_pairs, filtered_cell_genmuts_pairs = [], []

        for cell_name, (mutation_counts, filtered_mutation_counts) in results:
            cell_genemuts_pairs.append((cell_name, mutation_counts))
            filtered_cell_genmuts_pairs.append(
                (cell_name, filtered_mutation_counts))

        return (
            self._make_mutation_counts_df(cell_genemuts_pairs),
            self._make_mutation_counts_df(filtered_cell_genmuts_pairs),
        )


@click.command()
@click.option("--processes", default=1,
              prompt="number of processes to use for computation", type=int)
@click.option("--cosmicdb", prompt="path to cosmic db file (.tsv)",
              required=True)
@click.option("--refgenome", prompt="path to reference genome (.gtf)",
              required=True)
@click.option("--outfile", prompt="path to output file (.csv)", required=True)
@click.argument("files", required=True, nargs=-1)
def count_mutations(processes, cosmicdb, refgenome, outfile, files):
    """ count total number of mutations in each sample """
    print("Beginning setup (this may take several minutes!)")

    print("Loading COSMIC database...")
    cosmic_df = pd.read_csv(cosmicdb, delimiter='\t')
    print("Loading reference genome...")
    refgenome_df = pd.read_csv(refgenome, delimiter='\t', header=None)

    print("Building genome trees...")
    mutation_counter = MutationCounter(cosmic_df, refgenome_df)
    print("Setup complete.")

    print("Finding mutations...")
    result_df, filtered_result_df = mutation_counter.find_mutation_counts(
        files, processes=processes)

    print("Writing file...")
    outfile = Path(outfile)

    result_df.to_csv(outfile)
    filtered_result_df.to_csv(
        outfile
            .with_name(outfile.stem + "_filtered")
            .with_suffix(outfile.suffix)
    )

    print("Done!")



import os
import subprocess

# Set input language USA unicode encoding setting
# Necessary because click assumes ascii input unless otherwise specified
# https://click.palletsprojects.com/en/7.x/python3/
unicode_usa = 'en_US.utf-8'
os.environ['LC_LANG'] = unicode_usa
os.environ["LC_ALL"] = unicode_usa

HOME = os.path.expanduser("~")


def sanitize_path(path):
    """Make sure path is absolute and user-expanded"""
    return os.path.abspath(os.path.expanduser(path))


def maybe_add_slash(path):
    """Add a final trailing slash if it wasn't there already"""
    with_trailing_slash = path if path.endswith('/') else path + '/'
    return with_trailing_slash


def decode(stderr_or_stdout):
    decoded = stderr_or_stdout.decode("utf-8").splitlines()
    lines = [x.strip() for x in decoded]
    return lines


def get_stdout_from_command(command):
    """Run a program on the command line, and save the stdout

    Parameters
    ----------
    command : list
        list of strings to submit to subprocess.run

    Returns
    -------
    lines : list
        Newline-separated strings from output of command
    """
    result = subprocess.run(command, stdout=subprocess.PIPE)
    lines = decode(result.stdout)
    return lines


def get_stdout_stderr_from_command(command):
    """Run a program on the command line, and save the stdout

    Parameters
    ----------
    command : list
        list of strings to submit to subprocess.run

    Returns
    -------
    lines : list
        Newline-separated strings from output of command
    """
    result = subprocess.run(command, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    stdout = decode(result.stdout)
    stderr = decode(result.stderr)
    return stdout, stderr



# -*- coding: utf-8 -*-

# Import modified 'os' module with LC_LANG set so click doesn't complain
from cerebra.os_utils import os  # noqa: F401
from functools import partial

import click

from cerebra.germline_filter import germline_filter
from cerebra.count_mutations import count_mutations
from cerebra.find_aa_mutations import find_aa_mutations

click.option = partial(click.option, show_default=True)

settings = dict(help_option_names=['-h', '--help'])


@click.group(options_metavar='',
             subcommand_metavar='<command>',
             context_settings=settings)
def cli():
    """
    high-throughput summarizing of vcf entries following a sequencing
    experiment
    """
    pass


cli.add_command(germline_filter, name='germline-filter')
cli.add_command(count_mutations, name='count-mutations')
cli.add_command(find_aa_mutations, name='find-aa-mutations')

if __name__ == "__main__":
    cli()



"""
VCF.py
Kamil Slowikowski
October 30, 2013
Read VCF files. Works with gzip compressed files and pandas.
Note: This module ignores the genotype columns because
      I didn't need them at the time of writing.
Read more about VCF:
    http://vcftools.sourceforge.net/specs.html
Usage example:
    >>> import VCF
    >>> variants = VCF.lines('file.vcf.gz')
    >>> print variants.next()['CHROM']
    1
Use the generator to avoid loading the entire file into memory:
    >>> for v in VCF.lines('file.vcf.gz'):
    ...     print v['REF'], v['ALT']
    ...     break
    A T
If your file is not too large, read it directly into a DataFrame:
    >>> df = VCF.dataframe('file.vcf.gz')
    >>> df.columns
    Index([u'CHROM', u'POS', u'ID', u'REF', u'ALT', u'QUAL', u'FILTER',
    u'INFO'], dtype=object)
If your file is *very small* and you want to access INFO fields as columns:
    >>> df = VCF.dataframe('file.vcf.gz', large=False)
    >>> df.columns
    Index([u'CHROM', u'POS', u'ID', u'REF', u'ALT', u'QUAL', u'FILTER',
    u'GENE_NAME', u'GENE_ID', u'AA_POS', u'AA_CHANGE'], dtype=object)
LICENSE
This is free and unencumbered software released into the public domain.
Anyone is free to copy, modify, publish, use, compile, sell, or
distribute this software, either in source code form or as a compiled
binary, for any purpose, commercial or non-commercial, and by any
means.
In jurisdictions that recognize copyright laws, the author or authors
of this software dedicate any and all copyright interest in the
software to the public domain. We make this dedication for the benefit
of the public at large and to the detriment of our heirs and
successors. We intend this dedication to be an overt act of
relinquishment in perpetuity of all present and future rights to this
software under copyright law.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
For more information, please refer to <http://unlicense.org/>
"""

import gzip
from collections import OrderedDict

import pandas as pd

VCF_HEADER = ['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO']


def dataframe(filename, large=True):
    """Open an optionally gzipped VCF file and return a pandas.DataFrame with
    each INFO field included as a column in the dataframe.
    Note: Using large=False with large VCF files. It will be painfully slow.
    :param filename:    An optionally gzipped VCF file.
    :param large:       Use this with large VCF files to skip the ## lines and
                        leave the INFO fields unseparated as a single column.
    """
    if large:
        # Set the proper argument if the file is compressed.
        comp = 'gzip' if filename.endswith('.gz') else None
        # Count how many comment lines should be skipped.
        comments = _count_comments(filename)
        # Return a simple DataFrame without splitting the INFO column.
        return pd.read_table(filename, compression=comp, skiprows=comments,
                             names=VCF_HEADER, usecols=range(8))

    # Each column is a list stored as a value in this dict. The keys for this
    # dict are the VCF column names and the keys in the INFO column.
    result = OrderedDict()
    # Parse each line in the VCF file into a dict.
    for i, line in enumerate(lines(filename)):
        for key in line.keys():
            # This key has not been seen yet, so set it to None for all
            # previous lines.
            if key not in result:
                result[key] = [None] * i
        # Ensure this row has some value for each column.
        for key in result.keys():
            result[key].append(line.get(key, None))

    return pd.DataFrame(result)


def lines(filename):
    """Open an optionally gzipped VCF file and generate an OrderedDict for
    each line.
    """
    fn_open = gzip.open if filename.endswith('.gz') else open

    with fn_open(filename) as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            else:
                yield parse(line)


def parse(line):
    """Parse a single VCF line and return an OrderedDict.
    """
    result = OrderedDict()

    fields = line.rstrip().split('\t')

    # Read the values in the first seven columns.
    for i, col in enumerate(VCF_HEADER[:7]):
        result[col] = _get_value(fields[i])

    # INFO field consists of "key1=value;key2=value;...".
    infos = fields[7].split(';')

    for i, info in enumerate(infos, 1):
        # info should be "key=value".
        try:
            key, value = info.split('=')
        # But sometimes it is just "value", so we'll make our own key.
        except ValueError:
            key = 'INFO{}'.format(i)
            value = info
        # Set the value to None if there is no value.
        result[key] = _get_value(value)

    return result


def _get_value(value):
    """Interpret null values and return ``None``. Return a list if the value
    contains a comma.
    """
    if not value or value in ['', '.', 'NA']:
        return None
    if ',' in value:
        return value.split(',')
    return value


def _count_comments(filename):
    """Count comment lines (those that start with "#") in an optionally
    gzipped file.
    :param filename:  An optionally gzipped file.
    """
    comments = 0
    fn_open = gzip.open if filename.endswith('.gz') else open
    with fn_open(filename) as fh:
        for line in fh:
            if line.startswith('#'):
                comments += 1
            else:
                break
    return comments



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.read()

test_requirements = [
    'pytest', 'coverage', "flake8"
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='cerebra',
    version='1.0.7',
    description="finds mutants in your scRNA-seq experiment",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Lincoln Harris",
    author_email='lincoln.harris@czbiohub.org',
    url='https://github.com/czbiohub/cerebra',
    packages=[
        'cerebra',
    ],
    package_dir={'cerebra':
                     'cerebra'},
    include_package_data=True,
    install_requires=["cython>=0.29.1",
                        "pytest", 
                        "flake8",
                        "coverage",
                        "click",
                        "setuptools>=38.6.0",
                        "tqdm",
                        "colorama",
                        "numpy",
                        "pandas",
                        "vcfpy",
                        "pathos",
                        "ncls",
                        "hgvs",
                        "pyfaidx"],
    license="MIT",
    zip_safe=False,
    keywords='cerebra',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    entry_points={
        'console_scripts': [
            'cerebra = cerebra.commandline:cli'
        ]
    },
    test_suite='tests',
    tests_require=test_requirements
)